/*
 *	3DShell
 *	Scene.cpp - Implements the scene, draws it & processes user input passed from parent.
 *	Also provides the OpenGL-relevant response to some Win32 events, which parent passes down, eg resizing the viewport.
 *
 *  Hardly a real scene graph, but readily extensible into one.
 *  For adding new groups, only three functions have to be adjusted; SetupSceneMembers, DrawSceneMembers and ProcessHit.
 *
 *	Iain C, DATEMARK
 */

// ********************************* INCLUDES **********************************************

#include "GeneralHeader.h"

// ********************************* CONSTRUCTORS/DESTRUCTORS **********************************************

CScene::CScene()
{
	// *** Setup naming system ***
	NumberGroups=0;

	// *** Set starting world position ***
	SceneRotX = -50.0f;
	SceneRotZ = 0.0f;
	AutoRotate = false;
	WasRotating = false;
	LockoutControls = false;
	IsTracking = false;

	// *** Set lights ***
	//LightAmbient[4]=		{ 0.5f, 0.5f, 0.5f, 1.0f };		// Ambient light
	LightAmbient[0] = 0.5f;
	LightAmbient[1] = 0.5f;
	LightAmbient[2] = 0.5f;
	LightAmbient[3] = 0.5f;
	//LightDiffuse[4]=		{ 1.0f, 1.0f, 1.0f, 1.0f };		// Diffuse light
	LightDiffuse[0] = 1.0f;
	LightDiffuse[1] = 1.0f;
	LightDiffuse[2] = 1.0f;
	LightDiffuse[3] = 1.0f;
	//LightPosition[4]=	{ 0.0f, 0.0f, 2.0f, 1.0f };		// Light position
	LightPosition[0] = 0.0f;
	LightPosition[1] = 0.0f;
	LightPosition[2] = 2.0f;
	LightPosition[3] = 1.0f;
}

CScene::~CScene()
{
}

// ********************************* METHODS **********************************************

int CScene::DrawSceneMembers()	// Draw objects in the scene: World position's already set for us by Draw().
{
	//cBackdrop.Draw();					// Draw the backdrop. This actually leaves cam where it put it so doesn't need to push/pop the matrix.
	cWharf.Draw();						// Draw the wharf. Ditto on the cam.
	
	glPushMatrix();						// don't mess up the world origin
		cTaskMan.Draw();				// Draw the task manager
	glPopMatrix();						// reset to world origin

	return TRUE;
}

int CScene::SetupSceneMembers()	// Allow members to do their init
{
	cWharf.PostGLInit(GetNewObjectGroupName("WHARF"));			// Assign a base name & let the floor do its init
	cTaskMan.PostGLInit(GetNewObjectGroupName("TASKMAN"));		// Assign a base name & let the task manager do its init
	cBackdrop.PostGLInit(GetNewObjectGroupName("BACKDROP"));	// Assign a base name & let the task manager do its init
	return TRUE;
}

int CScene::ProcessHit(int objectName)
{
	// Work out which module it belongs to
	if (ObjectIsInGroup(objectName, "WHARF")) cWharf.ProcessHit(objectName);
	if (ObjectIsInGroup(objectName, "TASKMAN")) cTaskMan.ProcessHit(objectName);
	return TRUE;
}

int CScene::ReSizeGLScene(GLsizei width, GLsizei height)// Resize And Initialize The GL Window
{
	if (height==0)										// Prevent A Divide By Zero By
	{
		height=1;										// Making Height Equal One
	}

	glViewport(0,0,width,height);						// Reset The Current Viewport

	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix

	AspectRatio = (GLfloat)width/(GLfloat)height;		// Calculate The Aspect Ratio Of The Window
	gluPerspective(45.0f,AspectRatio,0.1f,100.0f);
	
	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix

	return TRUE;
}

int CScene::ProcessMouseMovement(int mx, int my, int omx, int omy)	// Process camera movement
{
	if (LockoutControls) return FALSE;

	// How much has it moved?
	int moveX = mx-omx;
	int moveY = my-omy;
	if (moveX!=0)	// Movement in the x: spin the Scene
		Spin((float)moveX);
	if (moveY!=0)	// Movement in the y: spin the Scene
		Tilt((float)moveY);
	
	return TRUE;
}

int CScene::Spin(float moveAmount)	// Spin world
{
	SceneRotZ = SceneRotZ + (moveAmount / 4);
	if (SceneRotZ > 359) SceneRotZ = 0;
	if (SceneRotZ < 0) SceneRotZ = 359;

	return TRUE;
}

int CScene::Tilt(float moveAmount)	// Tilt world
{
	SceneRotX = SceneRotX + (moveAmount / 4);
	if (SceneRotX > 0) SceneRotX = 0;
	if (SceneRotX < -85) SceneRotX = -85;

	return TRUE;
}

int CScene::PostGLInit()								// Do post-GL init
{
	glEnable(GL_LIGHTING);								// Enable lighting
	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);		// Setup The Ambient Light
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);		// Setup The Diffuse Light
	glLightfv(GL_LIGHT1, GL_POSITION,LightPosition);	// Position The Light
	glEnable(GL_LIGHT1);								// Enable Light One

	glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
	glClearColor(0.282f, 0.306f, 0.435f, 0.5f);			// Background

	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
	
	glDepthFunc(GL_LEQUAL);								// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations

	SetupSceneMembers();								// setup world objects

	return TRUE;
}

int CScene::Draw()	// Shift world (locate camera) & draw the Scene.
{
	if (IsTracking) DoCamTrack();						// Continue track sequence if necessary
	if (AutoRotate) Spin(1);							// Continue autorotate if necessary

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
	glLoadIdentity();									// Reset The Current Modelview Matrix

	glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);	// Position the Light

	SceneDistance=50 + (SceneRotX/7);
	glTranslatef(0.0f,0.0f,-SceneDistance);				// Move world into the screen
	glRotatef(SceneRotX,1.0f,0.0f,0.0f);				// Tilt forward/backward
	glRotatef(SceneRotZ,0.0f,0.0f,1.0f);				// Finally, rotate the world in its z-plane.

	glMatrixMode (GL_MODELVIEW);

	DrawSceneMembers();									// Draw objects in scene
	return TRUE;										// Keep Going
}

int CScene::FindHit (int mx, int my)	// Locate a mouse hit and tell the appropriate object group
{
	GLuint selectBuf [MAX_OBJECTS * 4];					// Our selection buffer
	GLint numberHits, i, j;								// For parsing the buffer
	int dummy,											// Info about hit object(s)
		maximumZ[MAX_OBJECTS],
		objectName[MAX_OBJECTS],
		closestObjectName,
		closestObjectZ;
	GLint viewport [4];									// Current Viewport
	
/*
	// FOR DEBUG
	char linez [255], temp[255], char text [255];
	int minimumZ
*/

	glGetIntegerv(GL_VIEWPORT, viewport);				// get current viewport
	glSelectBuffer(512, selectBuf);						// assign a selection buffer

	glMatrixMode(GL_PROJECTION);						// change matrix mode to projection
	
	glPushMatrix ();									// don't disturb things
		glLoadIdentity ();

		glRenderMode(GL_SELECT);						// switch to selection mode
		gluPickMatrix(mx, viewport[3]-my, 1, 1, viewport);	// Make virtual clipping region with center x, y and width and height 1 each w.r.t the current viewport.
		gluPerspective(45.0f,AspectRatio,0.1f,100.0f);	// set up a perspective projection matrix to 45 deg.
		
		glInitNames();									// init name stack
		glPushName(BASENAME_DUMMY);						// we don't want an empty one because we may try popping it in a sec & that's baaaaad
		
		glMatrixMode (GL_MODELVIEW);

		DrawSceneMembers();								// Call our render funtion to load object with names. Note that since our render mode is GL_SELECT nothing will be drawn on the screen due to this function call!

		glPopName();									// pop our dummy name

		numberHits = glRenderMode(GL_RENDER);			// When switching back glRenderMode returns no. of hits
		glMatrixMode(GL_PROJECTION);
	glPopMatrix ();

	glMatrixMode(GL_MODELVIEW);							// Change Matrix Mode to Model View

	if (numberHits > 0)									// We have a hit
	{
		// Parse selection buffer
		j = 0;
		for (i = 0; i < numberHits; i++)	// Step through objects hit
		{
			dummy			= selectBuf [j]; j++;
			dummy			= selectBuf [j]; j++;
			maximumZ[i]		= selectBuf [j]; j++;	// Store max z for this object
			objectName[i]	= selectBuf [j]; j++;	// Store name for this object
		}  

		// Find closest hit object
		closestObjectName = 0;
		closestObjectZ = MIN_Z;
		for (i = 0; i < numberHits; i++)	// Step through objects hit
		{
			if (maximumZ[i] > closestObjectZ)	// This object is closer than the one we're storing, update
			{
				closestObjectName = objectName[i];
				closestObjectZ = maximumZ[i];
			}
		}

		ProcessHit(closestObjectName);	// Process the hit
	}

	return TRUE;
}

int CScene::GetNewObjectGroupName(char *HumanReadableGroupName)
{
	int NextObjectGroupName;

	NextObjectGroupName = (NumberGroups+1) * OBJECT_GROUP_OFFSET;			// Base OpenGL name for this group
	
	sprintf (GroupNameLookupHRGN[NumberGroups], HumanReadableGroupName);	// Add to lookup tables
	GroupNameLookupGN[NumberGroups] = NextObjectGroupName;

	NumberGroups++;															// Increment group counter

	return NextObjectGroupName;
}

bool CScene::ObjectIsInGroup(int ObjectName, char *HumanReadableGroupName)
{
	int GroupBaseName=-1;
	int i;

	// Get bounds of HRGN
	for (i=0; i<NumberGroups; i++)
	{
		if (strcmp(GroupNameLookupHRGN[i], HumanReadableGroupName)==0)
		{
			GroupBaseName = GroupNameLookupGN[i];
		}
	}
	
	if (GroupBaseName==-1) return false;	// No such group name!
	
	if (ObjectName >= GroupBaseName && ObjectName < (GroupBaseName + OBJECT_GROUP_OFFSET))	// It's in there
	{
		return true;
	} else {																				// No it's not
		return false;
	}
}

int CScene::PauseRotation()	// Pause world auto-rotate
{
	WasRotating = AutoRotate;
	AutoRotate = false;
	return TRUE;
}

int CScene::ToggleRotation()	// Toggle world auto-rotate
{
	AutoRotate = !AutoRotate;
	return TRUE;
}

int CScene::ContinueRotation()	// Continue world auto-rotate
{
	AutoRotate = WasRotating;
	return TRUE;
}

int CScene::SetCamTrackPath(int PathToTrack)	// Lockout camera controls while the camera tracks to a pre-defined viewpoint
{
	GLfloat TargetSceneRotX, TargetSceneRotZ, CurSceneRotX, CurSceneRotZ, StepX, StepZ;

	CurSceneRotX = SceneRotX;	// Get local copies of scene co-ords
	CurSceneRotZ = SceneRotZ;

	switch(PathToTrack)		// Set target values
	{
		case CAMPATH_TRACK_TO_WHARF:
		{
			TargetSceneRotX = 0.0f;
			TargetSceneRotZ = 0.0f;
		}

		case CAMPATH_TRACK_TO_TASKMAN:
		{
			TargetSceneRotX = -85.0f;
			TargetSceneRotZ = 0.0f;
		}	
	}

	// Set up interpolation array for sequence
	         
	StepX = (TargetSceneRotX - CurSceneRotX) / CAMPATH_TRACK_OVER_FRAMES;
	StepZ = (TargetSceneRotZ - CurSceneRotZ) / CAMPATH_TRACK_OVER_FRAMES;

	for (int i=0; i<CAMPATH_TRACK_OVER_FRAMES; i++)
	{
		TrackingPosX[i]=CurSceneRotX;
		TrackingPosZ[i]=CurSceneRotZ;
		CurSceneRotX += StepX;
		CurSceneRotZ += StepZ;
	}

	LockoutControls = true;				// Lock out controls
	PauseRotation();					// Pause autorotate
	TrackFrame = 0;						// Set to first sequence frame
	IsTracking = true;					// Start tracking

	return TRUE;
}

int CScene::DoCamTrack()
{
	SceneRotX = TrackingPosX[TrackFrame];
	SceneRotZ = TrackingPosZ[TrackFrame];

	TrackFrame++;

	if (TrackFrame == CAMPATH_TRACK_OVER_FRAMES)	// It's over, turn controls back on & stop calling this function
	{
		IsTracking = false;
		LockoutControls = false;
	}

	return TRUE;
}
